/*********************************************************************
    This file is part of Siworin. Copyright © 2021 by Ruud Harmsen.

    Siworin is a simple word indexer and local search engine. See
    https://rudhar.com/sfreview/siworin/#siw04 and
    https://rudhar.com/sfreview/siworin/toolsrc/siworin .

    Siworin is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    Siworin is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Siworin.  If not, see <https://www.gnu.org/licenses/>.
  ********************************************************************/

/* Siworin - simple word indexer
   2, 16-20, 26 July, 12, 15, 17, 18 August 2021,
   5, 6, 13, 16, 17, 18 September 2021,
   21 October 2021,
   by Ruud Harmsen.
 */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "siworin.h"
#include "siworin-srchind.h"
#include "siworin-files.h"
#include "siworin-config.h"
#include "siworin-match.h"

static FILE *tmpfp;
static int siworin_compar (const void *key, const void *elt);
static int siworin_calcnumwords (void);
static char *modify_wrd (char *srch_wrd);

/* Pointer to non-existent array. As long as the pointers
   to array elements are not dereferenced, nothing goes
   wrong. See http://rudhar.com/sfreview/siworin/siworin12.htm
   for a much longer explanation. */
//static char *siworin_fakebase;
static char siworin_fakebase[] = "";

void siworin_handle_srchwrds (struct siworin_formdata_s *p_formdata)
{
   tmpfp = tmpfile();
   if (!tmpfp)
      SIWORIN_FAIL("tmpfile");

   /* Safety first, although this is supposed to have been
      called earlier on, from siworin-cgi.c */
   if (!p_formdata->words_pointers)
      siworin_index_srchwrds(p_formdata);

   if (p_formdata->regexp)
   {
      char *cmd;
      char *modified_wrd;
      FILE *pip;

      if (!(cmd = malloc(
            strlen(p_formdata->words_pointers[0]) +
            strlen(siworin_config_value(SIWORIN_CONFIG_INDEX_FILES_PREFIX))
            + 50 + sizeof SIWORIN_FORMAT3 + sizeof SIWORIN_FORMAT5)))
         SIWORIN_FAIL("fopens");

      if (!(modified_wrd = modify_wrd(p_formdata->words_pointers[0])))
         SIWORIN_FAIL("modify_wrd");

      sprintf(cmd,
         /* To get debugging material for siworin-match module test */
//       "LC_ALL=en_US.UTF-8 grep -E -m 1000 \'%s\' %s.words | tee /home/rudhar/mint",
         "LC_ALL=en_US.UTF-8 grep -E -m 1000 \'%s\' %s.words",
         modified_wrd,
         siworin_config_value(SIWORIN_CONFIG_INDEX_FILES_PREFIX));
      siworin_findlocsfromfiles(pip = popen(cmd, "r"), 0);
      free(cmd);
      free(modified_wrd);
      pclose(pip);
   }
   else
   {
      long numentries = siworin_calcnumwords();

      for (int i = 0; i < p_formdata->numwords; i++)
      {
         /* About siworin_fakebase, see
            https://rudhar.com/sfreview/siworin/siworin12.htm */
         (void)bsearch(p_formdata->words_pointers[i], siworin_fakebase, numentries,
            sizeof (*siworin_fakebase), siworin_compar);
      }
      rewind(tmpfp);
      siworin_findlocsfromfiles(tmpfp, p_formdata->combin);
   }

   siworin_free_index(p_formdata);
   fclose(tmpfp);
}

void siworin_index_srchwrds (struct siworin_formdata_s *p_formdata)
{
   char *cp;
   int inword = 0;

   /* Build a table of pointers to null-terminated search words */
   if (!(p_formdata->words_sepd_by_nulls =
         malloc(strlen(p_formdata->searchstring) + 1)))
      SIWORIN_FAIL("malloc");
   strcpy(p_formdata->words_sepd_by_nulls, p_formdata->searchstring);
   for (cp = p_formdata->words_sepd_by_nulls; *cp; cp++)
   {
      if (isspace(*cp))
      {
         *cp = '\0';
         inword = 0;
      }
      else if (!inword)
      {
         inword = 1;
         if (!(p_formdata->words_pointers =
            realloc(p_formdata->words_pointers,
               ++(p_formdata->numwords) * sizeof *(p_formdata->words_pointers))))
            SIWORIN_FAIL("realloc");
         p_formdata->words_pointers[p_formdata->numwords - 1] = cp;
      }
   }
   if (p_formdata->numwords > 1)
      p_formdata->regexp = 0;
   if (p_formdata->regexp)
      p_formdata->combin = 0;
}

void siworin_free_index (struct siworin_formdata_s *p_formdata)
{
   free(p_formdata->words_sepd_by_nulls);
   p_formdata->words_sepd_by_nulls = NULL;

   free(p_formdata->words_pointers);
   p_formdata->words_pointers = NULL;
}

static int siworin_compar (const void *key, const void *elt)
{
   size_t index = (char *)elt - siworin_fakebase;
   char bytepos_wbuf[SIWORIN_OFFSET_HEXDIGITS + 1 /* newline */ + 1 /* nullbyte */];
   unsigned long bytepos_w;
   char fndwrd[128], *cp;
   int compared;

   if (fseek(siworin_fp_wordoffs,
             index * (SIWORIN_OFFSET_HEXDIGITS + 1 /* newline */), SEEK_SET) != 0)
      SIWORIN_FAIL("fseek");
   if (fgets(bytepos_wbuf, sizeof bytepos_wbuf, siworin_fp_wordoffs) == NULL)
      SIWORIN_FAIL("fgets");

   sscanf(bytepos_wbuf, "%lx", &bytepos_w);

   if (fseek(siworin_fp_words, bytepos_w, SEEK_SET) != 0)
      SIWORIN_FAIL("fseek");
   if (fgets(fndwrd, sizeof fndwrd, siworin_fp_words) == NULL)
      SIWORIN_FAIL("fgets");
   if ((cp = strchr(fndwrd, SIWORIN_SEP)) != NULL)
      *cp = '\0';

   compared = strcmp((char *)key, fndwrd);

   if (compared == 0)
   {
      int c;

      /* The exact search word ("key") was found. The corresponding
         line in the words file (default: siworin.words) now needs
         to be written to a temp file, so function
         siworin_findlocsfromfiles can process that line, in a way
         that is compatible with processing multiple lines obtained
         from egrep, which was activated via a call to popen(3). */

      /* Reposition to where we were before the previous read. */
      if (fseek(siworin_fp_words, bytepos_w, SEEK_SET) != 0)
         SIWORIN_FAIL("fseek");

      /* Copy the line to a temp file, to be picked up by
         function siworin_findlocsfromfiles */
      fseek(tmpfp, 0, SEEK_END);
      while ((c = getc(siworin_fp_words)) != EOF)
      {
         putc(c, tmpfp);
         if (c == '\n')
            break;
      }
   }
   return compared;
}

static int siworin_calcnumwords (void)
{
   struct stat st;

   if (!siworin_fp_wordoffs)
      SIWORIN_FAIL("failed to call siworin_fileopen* functions");

   if (fstat(fileno(siworin_fp_wordoffs), &st) != 0)
      SIWORIN_FAIL("fstat");

   return st.st_size / (SIWORIN_OFFSET_HEXDIGITS + 1);
}

/* 21 October 2021:
   Without this modification of the search pattern, egrep would
   search the whole of each line of the file siworin.words . But
   much of that is useless, because a line contains a word, and
   the encoded info of where to find it. To restrict the search
   to just the word, we modify the beginning and end of pattern,
   when necessary.
   From tests, it becomes clear that this does little if anything
   for performance. But it does avoid senseless matches on hex
   offsets, and the users can now themselves use ^ and $ to
   indicate the start and end of the search word. */
static char *modify_wrd (char *srch_wrd)
{
   int caret_present, dollar_present;
   int len = strlen(srch_wrd);
   char *modified =
      malloc(len + 1 + sizeof SIWORIN_FORMAT3 + sizeof SIWORIN_FORMAT5);

   if (!modified)
      return modified;

   caret_present  = srch_wrd[0] == '^';
   dollar_present = srch_wrd[len - 1] == '$' &&
      !(len >= 2 && srch_wrd[len - 2] == '\\');

   sprintf(modified, "%s%.*s%s",
      caret_present ? "" : SIWORIN_FORMAT3,
      dollar_present ? len - 1 : len,
      srch_wrd,
      dollar_present ? SIWORIN_FORMAT4 : SIWORIN_FORMAT5);

#ifdef MAIN
   fprintf(stdout,
      "Caret present %d, dollar present %d. Turned\n%s\ninto\n%s\n\n",
      caret_present, dollar_present, srch_wrd, modified);
#endif

   return modified;
}

#include <wctype.h>
void lowercase_searchstring (char *pat)
{
   wchar_t *wcp;
   wchar_t *wpat = malloc(sizeof *wpat * (strlen(pat) + 1));

   if (!wpat)
      SIWORIN_FAIL("fopens");

   mbstowcs(wpat, pat, strlen(pat));
   for (wcp = wpat; *wcp; wcp++)
      *wcp = towlower(*wcp);
   wcstombs(pat, wpat, strlen(pat));
   free(wpat);
}

#ifdef MAIN
/* Testdriver, to see results in a shell, using a commandline program. */
static struct siworin_formdata_s formdata;
int main (int argc, char **argv)
{
   int number_of_arguments_handled = siworin_args_to_config_path(argv);

   if (siworin_get_explicit_config_path())
      argc -= number_of_arguments_handled,
      argv += number_of_arguments_handled;

   if (siworin_fileopen_paths(
       siworin_config_value(SIWORIN_CONFIG_INDEX_FILES_PREFIX),
       "r", "r"))
      SIWORIN_FAIL("fopens");

   if (siworin_fileopen_words(
       siworin_config_value(SIWORIN_CONFIG_INDEX_FILES_PREFIX),
       "r", "r"))
      SIWORIN_FAIL("fopens");

   if (argc == 3 && argv[2] && strcmp(argv[2], "grep") == 0)
   {
      formdata.regexp = 1;
      formdata.combin = 0;
   }
   else
   {
      formdata.regexp = 0;
      formdata.combin = 1;
   }

   if (argc >= 2)
   {
      strncpy(formdata.searchstring, argv[1], sizeof formdata.searchstring);
      lowercase_searchstring(formdata.searchstring);
      siworin_handle_srchwrds(&formdata);
   }
   siworin_fileopen_close();

   return 0;
}
#endif
